/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2007-2008 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */

/*!
 *  \file   toppers/csv.hpp
 *  \brief  CSVt@Co͂Ɋւ錾`
 */
#ifndef TOPPERS_CSV_HPP_
#define TOPPERS_CSV_HPP_

#include <stdexcept>
#include <ostream>
#include <iterator>
#include <string>
#include <vector>
#include <algorithm>
#include <boost/utility.hpp>
#include "toppers/misc.hpp"

namespace toppers
{

  class csv_error : public std::runtime_error
  {
  public:
    explicit csv_error( long line )
      : std::runtime_error( "CSV parse error" ), line_( line )
    {
    }
    long line() const
    {
      return line_;
    }
  private:
    long line_;
  };

  /*!
   *  \class  basic_csv csv.hpp "toppers/csv.hpp"
   *  \brief  CSV t@C̉̓NX
   */
  template < typename CharT, class Traits = std::char_traits< CharT > >
    class basic_csv
  {
  public:
    typedef std::basic_string< CharT, Traits > string_type;
    typedef typename std::vector< std::vector< string_type > >::const_iterator const_iterator;
    typedef typename std::vector< std::vector< string_type > >::const_reference const_reference;
    typedef typename std::vector< std::vector< string_type > >::value_type value_type;
    typedef typename std::vector< std::vector< string_type > >::size_type size_type;

    basic_csv();
    template < class InputIterator >
      basic_csv( InputIterator first, InputIterator last )
    {
      read( first, last );
    }

    template < class InputIterator >
      void read( InputIterator first, InputIterator last )
    {
      const CharT dquo = widen< CharT >( '\"' );
      const CharT comma = widen< CharT >( ',' );
      const CharT lf = widen< CharT >( '\n' );
      enum
      {
        none,
        non_escaped,
        escape_opened,
        escape_closed
      } state = none;
      std::vector< std::vector< string_type > > records;
      std::vector< string_type > record;
      string_type field;
      long line = 1;

      record.reserve( 16 );
      field.reserve( 255 );

      while ( first != last )
      {
        CharT ch = *first;
        if ( ch == dquo )
        {
          switch ( state )
          {
          case none:
            state = escape_opened;
            break;
          case escape_opened:
            ++first;
            ch = *first;
            if ( ch == dquo ) // "" ̂ŁA" ǉ
            {
              field.push_back( dquo );
            }
            else  //  "
            {
              state = escape_closed;
            }
            continue;   // ++first ȏsȂ
          default:
            throw csv_error( line );
            break;
          }
        }
        else  // ch != dquo
        {
          if ( state == escape_opened )
          {
            field.push_back( ch );
          }
          else
          {
            if ( ch == comma || ch == lf )
            {
              record.push_back( field );
              field.clear();
              state = none;
              if ( ch == lf ) // siRFC4180 ł͌ɂ CR-LF Ȃ΂ȂȂAƂ肠 '\n' Ƃj
              {
                records.push_back( record );
                record.clear();
                ++line;
              }
            }
            else
            {
              field.push_back( ch );
//              CharT const* debug_string = field.c_str();
            }
          }
        }
        ++first;
      }
      records_.swap( records );
    }
    template < class OutputIterator >
      void write( OutputIterator result ) const
    {
      const CharT dquo = widen< CharT >( '\"' );
      const CharT comma = widen< CharT >( ',' );
      const CharT lf = widen< CharT >( '\n' );

      for ( typename std::vector< std::vector< string_type > >::const_iterator r_iter( records_.begin() ), r_last( records_.end() );
            r_iter != r_last;
            ++r_iter )
      {
        for ( typename std::vector< string_type >::const_iterator f_iter( r_iter->begin() ), f_last( r_iter->end() );
              f_iter != f_last;
              ++f_iter )
        {
          std::string field;
          bool need_escape = false;

          for ( typename string_type::const_iterator s_iter( f_iter->begin() ), s_last( f_iter->end() );
                s_iter != s_last;
                ++s_iter )
          {
            CharT ch = *s_iter;
            if ( ch == dquo )
            {
              field.push_back( dquo );
              need_escape = true;
            }
            else if ( ch == comma || ch == lf )
            {
              need_escape = true;
            }
            field.push_back( ch );
          }

          if ( need_escape )
          {
            field = dquo + field + dquo;
          }
          result = std::copy( field.begin(), field.end(), result );
          if ( boost::next( f_iter ) != f_last )
          {
            *result++ = comma;
          }
        }
        *result++ = lf;
      }
    }

    const_iterator begin() const
    {
      return records_.begin();
    }
    const_iterator end() const
    {
      return records_.end();
    }
    const_reference at( size_type pos ) const
    {
      return records_.at( pos );
    }
    const_reference operator[]( size_type pos ) const
    {
      return records_[ pos ];
    }
    bool empty() const
    {
      return records_.empty();
    }
    size_type size() const
    {
      return records_.size();
    }
    void swap( basic_csv& other )
    {
      records_.swap( other.records_ );
    }
  private:
    std::vector< std::vector< string_type > > records_;
  };

  typedef basic_csv< char > csv;
  typedef basic_csv< wchar_t > wcsv;

}

#endif  // TOPPERS_CSV_HPP_
